@@ -10,26 +10,24 @@ local attr_dir = luatexbase.attributes['ltj@dir']
 local attr_icflag = luatexbase.attributes['ltj@icflag']
 
 local cat_lp = luatexbase.catcodetables['latex-package']
-local Dnode = node.direct or node
-local nullfunc = function (n) return n end
-local to_node = (Dnode ~= node) and Dnode.tonode or nullfunc
-local to_direct = (Dnode ~= node) and Dnode.todirect or nullfunc
-local has_attr = Dnode.has_attribute
-local set_attr = Dnode.set_attribute
-local insert_before = Dnode.insert_before
-local insert_after = Dnode.insert_after
-local getid = (Dnode ~= node) and Dnode.getid or function(n) return n.id end
-local getsubtype = (Dnode ~= node) and Dnode.getsubtype or function(n) return n.subtype end
-local getlist = (Dnode ~= node) and Dnode.getlist or function(n) return n.head end
-local setfield = (Dnode ~= node) and Dnode.setfield or function(n, i, c) n[i] = c end
-local getfield = (Dnode ~= node) and Dnode.getfield or function(n, i) return n[i] end
-local node_new = Dnode.new
-local node_tail = Dnode.tail
-local node_free = Dnode.free
-local node_remove = Dnode.remove
-local node_next = (Dnode ~= node) and Dnode.getnext or node.next
-local traverse = Dnode.traverse
-local traverse_id = Dnode.traverse_id
+local to_node = node.direct.tonode
+local to_direct = node.direct.todirect
+local has_attr = node.direct.has_attribute
+local set_attr = node.direct.set_attribute
+local insert_before = node.direct.insert_before
+local insert_after = node.direct.insert_after
+local getid = node.direct.getid
+local getsubtype = node.direct.getsubtype
+local getlist = node.direct.getlist
+local setfield = node.direct.setfield
+local getfield = node.direct.getfield
+local node_new = node.direct.new
+local node_tail = node.direct.tail
+local node_free = node.direct.free
+local node_remove = node.direct.remove
+local node_next = node.direct.getnext
+local traverse = node.direct.traverse
+local traverse_id = node.direct.traverse_id
 local start_time_measure, stop_time_measure
    = ltjb.start_time_measure, ltjb.stop_time_measure
 local abs = math.abs
@@ -65,7 +63,7 @@ local page_direction
 --
 local dir_pool
 do
-   local node_copy = Dnode.copy
+   local node_copy = node.direct.copy
    dir_pool = {}
    for _,i in pairs({dir_tate, dir_yoko, dir_dtou, dir_utod}) do
       local w = node_new(id_whatsit, sid_user)
@@ -108,14 +106,14 @@ do
       end
    end
    function get_dir_count()
-      for i=tex_nest.ptr, 1, -1 do
-	 local h = tex_nest[i].head.next
-	 if h then
-	    local t = get_dir_count_inner(h)
-	    if t~=0 then return t end
-	 end
-      end
-      return page_direction
+       for i=tex_nest.ptr, 1, -1 do
+	   local h = tex_nest[i].head.next
+	   if h then
+	       local t = get_dir_count_inner(h)
+	       if t~=0 then return t end
+	   end
+       end
+       return page_direction
    end
    function get_adjust_dir_count()
       for i=tex_nest.ptr, 1, -1 do
@@ -140,7 +138,7 @@ do
    local node_traverse = node.traverse
    local STCK = luatexja.userid_table.STCK
    local IHB = luatexja.userid_table.IHB
-   local sid_local = node.subtype('local_par')
+   local id_local = node.id('local_par')
 
    local function test_list(h, lv)
       if not h then
@@ -158,10 +156,8 @@ do
 		  elseif not(uid==IHB or uid==STCK) then
 		     flag = 0; break -- error
 		  end
-	       elseif ps~=sid_local then
-		  flag = 0; break
 	       end
-	    else
+	    elseif p.id~=id_local then
 	       flag = 0; break
 	    end
 	 end
@@ -172,6 +168,8 @@ do
 	 end
       end
    end
+   local node_next_node, node_tail_node = node.next, node.tail
+   local insert_after_node = node.insert_after
    function luatexja.direction.set_list_direction_hook(v)
       local lv = tex_nest.ptr -- must be >= 1
       if not v then
@@ -182,10 +180,12 @@ do
       elseif v=='adj' then
          v = get_adjust_dir_count()
       end
-      local h = to_direct(tex_nest[lv].head)
-      local w = dir_pool[v]()
-      insert_after(h, h, w)
-      tex_nest[lv].tail = to_node(node_tail(w))
+      local h = tex_nest[lv].head
+      local hn = node.next(h)
+      hn = (hn and hn.id==id_local) and hn or h
+      local w = to_node(dir_pool[v]())
+      insert_after_node(h, hn, w)
+      tex_nest[lv].tail = node_tail_node(w)
       ensure_tex_attr(attr_icflag, 0)
       ensure_tex_attr(attr_dir, 0)
    end
@@ -200,14 +200,21 @@ do
       elseif v=='adj' then
          v,name = get_adjust_dir_count(), nil
       end
+      local current_nest = tex_nest[lv]
       if tex.currentgrouptype==6 then
 	 ltjb.package_error(
                  'luatexja',
                  "You can't use `\\" .. name .. "' in an align",
-		 "To change direction in an align, \n"
-		    .. "you shold use \\hbox or \\vbox.")
+		 "To change the direction in an align, \n"
+		 .. "you shold use \\hbox or \\vbox.")
+      elseif current_nest.mode == ltjs.hmode or abs(current_nest.mode) == ltjs.mmode then
+	 ltjb.package_error(
+                 'luatexja',
+		 "Improper `\\" .. name .. "'",
+		 'You cannot change the direction in unrestricted horizontal mode \n'
+		 .. 'nor math modes.')
       else
-	 local h = (lv==0) and tex.lists.page_head or tex_nest[lv].head.next
+	 local h = (lv==0) and tex.lists.page_head or current_nest.head.next
 	 local flag,w = test_list(h,lv)
 	 if flag==0 then
 	    if lv==0 and not page_direction then
@@ -222,8 +229,15 @@ do
 	 elseif flag==1 then
 	    node_set_attr(w, attr_dir, v)
 	    if lv==0 then page_direction = v end
-	 else
-	    if lv==0 then page_direction = v end
+	 elseif lv==0 then
+	    page_direction = v
+	 else -- flag == 2: need to create dir whatsit.
+	    local h = current_nest.head
+	    local hn = node.next(h)
+	    hn = (hn and hn.id==id_local) and hn or h
+	    local w = to_node(dir_pool[v]())
+	    insert_after_node(h,hn,w)
+	    current_nest.tail = node_tail_node(w)
 	 end
          ensure_tex_attr(attr_icflag, 0)
       end
@@ -238,7 +252,7 @@ local function create_dir_whatsit(hd, gc, new_dir)
 	    getsubtype(hd)==sid_user and getfield(hd, 'user_id')==DIR then
       set_attr(hd, attr_icflag,
 	       get_attr_icflag(hd) + PROCESSED_BEGIN_FLAG)
-      local n = node_next(hd)
+      local n =node_next(hd)
       if n then
 	 set_attr(n, attr_icflag,
 		  get_attr_icflag(n) + PROCESSED_BEGIN_FLAG)
@@ -259,28 +273,35 @@ end
 
 -- hpack_filter, vpack_filter, post_line_break_filter
 -- の結果を組方向を明示するため，先頭に dir_node を設置
+local get_box_dir
 do
    local function create_dir_whatsit_hpack(h, gc)
       local hd = to_direct(h)
-      if gc=='fin_row' or gc == 'preamble'  then
+      if gc=='fin_row' then
 	 if hd  then
+	    for p in traverse_id(15, hd) do -- unset
+	       if get_box_dir(p, 0)==0 then
+                  setfield(p, 'head', create_dir_whatsit(getlist(p), 'fin_row', ltjs.list_dir))
+               end
+	    end
 	    set_attr(hd, attr_icflag, PROCESSED_BEGIN_FLAG)
 	    ensure_tex_attr(attr_icflag, 0)
 	 end
 	 return h
+      elseif gc == 'preamble'  then
       else
 	 adjust_badness(hd)
 	 return to_node(create_dir_whatsit(hd, gc, ltjs.list_dir))
       end
    end
 
-   luatexbase.add_to_callback('hpack_filter',
+   ltjb.add_to_callback('hpack_filter',
 			      create_dir_whatsit_hpack, 'ltj.create_dir_whatsit', 10000)
 end
 
 do
    local function create_dir_whatsit_parbox(h, gc)
-      stop_time_measure('tex_linebreak')
+      stop_time_measure('tex_linebreak');
       -- start 側は ltj-debug.lua に
       local new_dir = ltjs.list_dir
       for line in traverse_id(id_hlist, to_direct(h)) do
@@ -289,7 +310,7 @@ do
       ensure_tex_attr(attr_dir, 0)
       return h
    end
-   luatexbase.add_to_callback('post_linebreak_filter',
+   ltjb.add_to_callback('post_linebreak_filter',
 			      create_dir_whatsit_parbox, 'ltj.create_dir_whatsit', 10000)
 end
 
@@ -483,16 +504,18 @@ end
 
 -- 1st ret val: b の組方向
 -- 2nd ret val はその DIR whatsit
-local function get_box_dir(b, default)
+function get_box_dir(b, default)
    start_time_measure('get_box_dir')
    local dir = has_attr(b, attr_dir) or 0
    local bh = getfield(b,'head')
    -- b は insert node となりうるので getlist() は使えない
    local c
-   for bh in traverse_id(id_whatsit, bh) do
-      if getsubtype(bh)==sid_user and getfield(bh, 'user_id')==DIR then
-	 c = bh
-    	 dir = (dir==0) and has_attr(bh, attr_dir) or dir
+   if bh~=0 then -- bh != nil
+      for bh in traverse_id(id_whatsit, bh) do
+         if getsubtype(bh)==sid_user and getfield(bh, 'user_id')==DIR then
+            c = bh
+           dir = (dir==0) and has_attr(bh, attr_dir) or dir
+         end
       end
    end
    -- for i=1,2 do
@@ -514,7 +537,7 @@ do
       start_time_measure('box_primitive_hook')
       local list_dir = get_dir_count()%dir_math_mod
       local b = getbox(tex_getcount('ltj@tempcnta'))
-      if b then
+      if b and getlist(to_direct(b)) then
 	 local box_dir = get_box_dir(to_direct(b), dir_yoko)
 	 if box_dir%dir_math_mod ~= list_dir then
 	    ltjb.package_error(
@@ -578,7 +601,7 @@ local function unwrap_dir_node(b, head, box_dir)
    end
    local shift_old, b_dir, wh = nil, get_box_dir(bh, 0)
    if wh then
-      Dnode.flush_list(getfield(wh, 'value'))
+      node.direct.flush_list(getfield(wh, 'value'))
       setfield(wh, 'value', nil)
    end
    return nh, nb, bh, b_dir
@@ -617,13 +640,16 @@ do
       local bh = getlist(b)
       local box_dir, dn =  get_box_dir(b, ltjs.list_dir)
       -- 既に b の中身にあるwhatsit
-
+      if (box_dir<dir_node_auto) and (not dn) then
+	bh = create_dir_whatsit(bh, 'make_dir_whatsit', dir_yoko)
+	dn = bh; setfield(b, 'head', bh)
+      end
       if box_dir%dir_math_mod==new_dir then
 	 if box_dir>=dir_node_auto then
 	    -- dir_node としてカプセル化されている
 	    local _, dnc = get_box_dir(b, 0)
 	    if dnc then -- free all other dir_node
-	       Dnode.flush_list(getfield(dnc, 'value'))
+	       node.direct.flush_list(getfield(dnc, 'value'))
 	       setfield(dnc, 'value', nil)
 	    end
 	    set_attr(b, attr_dir, box_dir%dir_math_mod + dir_node_auto)
@@ -653,7 +679,7 @@ do
 	       db=x; break
 	    end
 	 end
-	 Dnode.flush_list(getfield(dn, 'value'))
+	 node.direct.flush_list(getfield(dn, 'value'))
 	 setfield(dn, 'value', nil)
 	 db = db or create_dir_node(b, box_dir, new_dir, false)
 	 local w = getfield(b, 'width')
@@ -684,7 +710,8 @@ do
    end
 
    -- lastbox
-   local node_prev = (Dnode~=node) and Dnode.getprev or node.prev
+   local node_prev = (node.direct~=node) and node.direct.getprev or node.prev
+   local id_glue = node.id('glue')
    local function lastbox_hook()
       start_time_measure('box_primitive_hook')
       local bn = tex_nest[tex_nest.ptr].tail
@@ -692,6 +719,14 @@ do
 	 local b, head = to_direct(bn), to_direct(tex_nest[tex_nest.ptr].head)
 	 local bid = getid(b)
 	 if bid==id_hlist or bid==id_vlist then
+            local p = getlist(b)
+	    -- alignment の各行の中身が入ったボックス
+            if p and getid(p)==id_glue and getsubtype(p)==12 then -- tabskip
+	       local np = node_next(p); local npid = getid(np)
+	       if npid==id_hlist or npid==id_vlist then
+		  setfield(b, 'head', create_dir_whatsit(p, 'align', get_box_dir(np, 0)))
+	       end
+            end
 	    local box_dir =  get_box_dir(b, 0)
 	    if box_dir>= dir_node_auto then -- unwrap dir_node
 	       local p = node_prev(b)
@@ -702,7 +737,7 @@ do
 	    end
 	    local _, wh =  get_box_dir(b, 0) -- clean dir_node attached to the box
 	    if wh then
-	       Dnode.flush_list(getfield('value', wh))
+	       node.direct.flush_list(getfield('value', wh))
 	       setfield(wh, 'value', nil)
 	    end
 	 end
@@ -763,8 +798,8 @@ do
    luatexja.direction.get_box_dim = get_box_dim
 
    -- return value: (changed dimen of box itself?)
-   local scan_dimen, scan_int = luatexja.token.scan_dimen, luatexja.token.scan_int
-   local scan_keyword = luatexja.token.scan_keyword
+   local scan_dimen, scan_int = token.scan_dimen, token.scan_int
+   local scan_keyword = token.scan_keyword
    local function set_box_dim_common(key, s, l_dir)
       local s_dir, wh = get_box_dir(s, dir_yoko)
       s_dir = s_dir%dir_math_mod
@@ -863,7 +898,7 @@ do
 end
 
 do
-   local getbox, setbox, copy_list = tex.getbox, tex.setbox, Dnode.copy_list
+   local getbox, setbox, copy_list = tex.getbox, tex.setbox, node.direct.copy_list
    -- raise, lower
    function luatexja.direction.raise_box()
       start_time_measure('box_primitive_hook')
@@ -910,7 +945,7 @@ do
 	       'luatexja',
 	       'Direction Incompatible',
 	       "\\vadjust's argument and outer vlist must have same direction.")
-	    Dnode.last_node()
+	    node.direct.last_node()
 	 end
       end
       stop_time_measure('box_primitive_hook')
@@ -926,6 +961,9 @@ do
       local list_dir = get_dir_count()
       local a = tex_nest[tex_nest.ptr].tail
       local ad = to_direct(a)
+      if (not a) or getid(ad)~=id_ins then
+	  a = node.tail(tex.lists.page_head); ad = to_direct(a)
+      end
       if a and getid(ad)==id_ins then
 	 local h = getfield(ad, 'head')
 	 if getid(h)==id_whatsit and
@@ -947,7 +985,25 @@ end
 
 -- vsplit
 do
-   local split_dir_whatsit
+   local split_dir_whatsit, split_dir_head
+   local cat_lp = luatexbase.catcodetables['latex-package']
+   local sprint, scan_int, tex_getbox = tex.sprint, token.scan_int, tex.getbox
+   function luatexja.direction.vsplit()
+      local n = scan_int();
+      local p = to_direct(tex_getbox(n))
+      split_dir_head = nil
+      if p then
+	 local bh = getlist(p)
+	 if getid(bh)==id_whatsit and getsubtype(bh)==sid_user and getfield(bh, 'user_id')==DIR 
+	    and node_next(bh) then
+	    ltjs.list_dir = has_attr(bh, attr_dir)
+	    local q = node_next(p)
+	    setfield(p, 'head', node_remove(bh,bh,bh))
+	    split_dir_head = bh
+	 end
+      end
+      sprint(cat_lp, '\\ltj@@orig@vsplit' .. tostring(n))
+   end	
    local function dir_adjust_vpack(h, gc)
       start_time_measure('direction_vpack')
       local hd = to_direct(h)
@@ -956,19 +1012,11 @@ do
 	 hd = create_dir_whatsit_vbox(hd, gc)
 	 split_dir_whatsit = hd
       elseif gc=='split_off'  then
-	 for  bh in traverse_id(id_whatsit, hd) do
-	    if getsubtype(bh)==sid_user and getfield(bh, 'user_id')==DIR then
-	       ltjs.list_dir  = has_attr(bh, attr_dir); break
-	    end
+	 if split_dir_head then
+	    list_dir = has_attr(split_dir_head, attr_dir)
+	    hd = insert_before(hd, hd, split_dir_head)
+	    split_dir_head=nil
 	 end
-	 -- local bh=hd
-	 -- for i=1,2 do
-	 --    if bh and getid(bh)==id_whatsit
-	 --    and getsubtype(bh)==sid_user and getfield(bh, 'user_id')==DIR then
-	 --       ltjs.list_dir  = has_attr(bh, attr_dir); break
-	 --    end
-	 --    bh = node_next(bh)
-	 -- end
 	 if split_dir_whatsit then
 	    -- adjust direction of 'split_keep'
 	    set_attr(split_dir_whatsit, attr_dir, ltjs.list_dir)
@@ -978,13 +1026,15 @@ do
 	 split_dir_whatsit=nil
       else
 	 adjust_badness(hd)
-	 hd = process_dir_node(create_dir_whatsit_vbox(hd, gc), gc)
+	 -- hd = process_dir_node(create_dir_whatsit_vbox(hd, gc), gc)
+	 -- done in append_to_vpack callback
+	 hd = create_dir_whatsit_vbox(hd, gc)
 	 split_dir_whatsit=nil
       end
       stop_time_measure('direction_vpack')
       return to_node(hd)
    end
-   luatexbase.add_to_callback('vpack_filter',
+   ltjb.add_to_callback('vpack_filter',
 			      dir_adjust_vpack,
 			      'ltj.direction', 10000)
 end
@@ -994,7 +1044,7 @@ do
    local function dir_adjust_pre_output(h, gc)
       return to_node(create_dir_whatsit_vbox(to_direct(h), gc))
    end
-   luatexbase.add_to_callback('pre_output_filter',
+   ltjb.add_to_callback('pre_output_filter',
 			      dir_adjust_pre_output,
 			      'ltj.direction', 10000)
 
@@ -1009,24 +1059,7 @@ do
    end
 end
 
--- buildpage filter
-do
-   local function dir_adjust_buildpage(info)
-      if info=='box' then
-	 local head = to_direct(tex.lists.contrib_head)
-	 if head then
-	    head = make_dir_whatsit(head,
-				  node_tail(head),
-				  get_dir_count(),
-				  'buildpage')
-	    tex.lists.contrib_head = to_node(head)
-	 end
-      end
-   end
-   luatexbase.add_to_callback('buildpage_filter',
-			      dir_adjust_buildpage,
-			      'ltj.direction', 10000)
-end
+-- append_to_vlist filter: done in ltj-lineskip.lua
 
 -- finalize (executed just before \shipout)
 -- we supply correct pdfsavematrix nodes etc. inside dir_node
@@ -1034,6 +1067,11 @@ do
    local finalize_inner
    local function finalize_dir_node(db,new_dir)
       local b = getlist(db)
+      if getid(b)==id_whatsit and getsubtype(b)==sid_user
+         and getfield(b, 'user_id')==DIR then
+         local ob = b; b = node_remove(b,b); setfield(db, 'head', b);
+	 node_free(ob)
+      end
       finalize_inner(b)
       local w = getfield(b, 'width')
       local h = getfield(b, 'height')
@@ -1042,11 +1080,12 @@ do
       local dn_h = getfield(db, 'height')
       local dn_d = getfield(db, 'depth')
       local db_head, db_tail
-      for _,v in ipairs(dir_node_aux
-			[get_box_dir(b, dir_yoko)%dir_math_mod][new_dir][getid(b)]) do
+      local t = dir_node_aux[get_box_dir(b, dir_yoko)%dir_math_mod][new_dir]
+      t = t and t[getid(b)]; if not t then return end
+      for _,v in ipairs(t) do
          local cmd, arg, nn = v[1], v[2]
          if cmd=='kern' then
-            nn = node_new(id_kern)
+            nn = node_new(id_kern, 1)
             setfield(nn, 'kern', arg(w, h, d, dn_w, dn_h, dn_d))
          elseif cmd=='whatsit' then
             nn = node_new(id_whatsit, arg)
@@ -1085,7 +1124,7 @@ do
       end
    end
    local getbox = tex.getbox
-   local setbox, copy = Dnode.setbox, Dnode.copy
+   local setbox, copy = node.direct.setbox, node.direct.copy
    local lua_mem_kb = 0
    function luatexja.direction.finalize()
       local a = to_direct(tex.getbox("AtBeginShipoutBox"))
@@ -1098,7 +1137,6 @@ do
       finalize_inner(shipout_temp)
       setbox('global', "AtBeginShipoutBox", copy(getlist(shipout_temp)))
       setfield(shipout_temp, 'head',nil)
-
       -- garbage collect
       --local m = collectgarbage('count')
       --if m>lua_mem_kb+20480 then
